"""
This module contains two classes:
	* Analyzer: a generic analyzer. It can be fed both from text strings and from files. \
		You can also store a representation of the state of the analyzer to be retrieved later, with the \
		``from_file`` class method or the ``load`` method.

	* EnglishAnalyzer: an special analyzer for the English language. 


"""
from freqens.normalized_counter import NormalizedCounter
from itertools import chain
import heapq, operator, json, os

here = os.path.dirname(os.path.abspath(__file__))
relative_path = lambda s: os.path.join(here, s)

def counter_distance(counter1, counter2):
	""" Euclidean distance on the frequency distribution space """
	keys = set( chain(counter1.elements(), counter2.elements()) ) 

	return sum( (counter1[key] - counter2[key])**2 for key in keys )

class Analyzer(object):
	""" 
	The class that performs the analysis. 
	You can feed an analyzer from different sources (strings, files... ) so that
	it extracts the target frequency distribution and ask
	it to score supplied content based on frequency similarity
	"""
	def __init__(self, content = None):
		""" Build an analyzer

		Content can be a string or a dict with absolute frequencies like in:
		`` Analyzer({"a": 4, "b": 8, "c": 1}) ``
		"""
		self.counter = NormalizedCounter(content)

	def feed(self, content):
		""" Feeds the analyzer with a string

		:param content: the string to be fed to the analyzer 
		"""
		self.counter.insert(content)

	def feed_from_raw_file(self, filename):
		""" Feeds the analyzer with the content of a file
			Every character will be taken into account, including newline chars.

			:param filename: the path of the file that will be fed to the analyzer
		"""
		with open(filename) as f:
			content = f.read()
			self.feed(content)

	def score(self, content):
		"""
		Assigns a score to any string. The smaller, the more similar frequency distribution. \
		0 means that the frequency distributions of both the content and the analyzer are equal.

		:param content: the string to be scored.

		:returns: a float number

		"""
		new_counter = NormalizedCounter()
		new_counter.insert(content)

		return counter_distance(self.counter, new_counter)

	def choose_best(self, strings, n=1):
		""" 
		Returns the n strings whose frequency distribution is most similar
		to the one fed to the analyzer.

		:param strings: an iterator with the strings where the Analyzer will looked for the 
			best strings.
		:param n: an integer specifying the number of strings which will be returned.

		:returns: an iterable containing the ``n`` best strings sorted by frequency similarity

		"""
		scores = { string: self.score(string) for string in strings }

		return map(operator.itemgetter(0), heapq.nsmallest(n, scores.iteritems(), operator.itemgetter(1)))

	def serialize(self):
		""" Returns a json representation of the analyzer 

		:returns: a string containing a json representation of the absolute frequencies
			the analyzer has been fed with."""
		content = self.counter.absolute_counts()

		return json.dumps(content)
 
	def store(self, filename):
		""" Stores the json representation of the analyzer to a file """
		with open(filename, "w") as f:
			f.write(self.serialize())

	def load(self, filename):
		""" Loads a frequency distribution file and adds it to the current distribution """
		with open(filename) as f:
			counter = NormalizedCounter(json.loads(f.read()))
			self.counter += counter


	def discard(self, chars):
		""" Removes the chars in chars from the counter

			:param chars: an interable consisting of the chars \
				whose frequency will be set to 0

		"""
		for char in chars:
			del self.counter[char]

	def transform_keys(self, transformation):
		""" Maps the keys to other new keys to get a new frequency distribution


			The relative frequency of keys that map to the same key will be added in 
			order to get the new frequency distribution.

			:param transformation: a callable object that maps chars to chars"""
		self.counter.transform(transformation)

	def keys(self):
		""" Returns the characters whose frequency is greater than 0 """
		return self.counter.elements()

	@classmethod
	def from_raw_file(self, filename):
		""" Returns an analyzer whose frequency distribution is read from the file content """
		analyzer = Analyzer()
		analyzer.feed_from_raw_file(filename)

		return analyzer

	@classmethod
	def from_file(self, filename):
		""" Reads a frequency distribution from a JSON file as stored by store method """
		analyzer = Analyzer()
		analyzer.load(filename)
		return analyzer
		

class EnglishAnalyzer(Analyzer):
	""" An analyzer for the english language """
	def __init__(self, blank_spaces=True, case_sensitive=True, just_alpha=False):
		super(EnglishAnalyzer, self).__init__()

		self.load(relative_path("data/english-export.txt"))

		if not blank_spaces:
			self.discard([" ", "\t"])

		self.case_sensitive = case_sensitive
		if not case_sensitive:
			self.transform_keys(lambda s: s.lower())

		if just_alpha:
			valid_symbol = lambda c: c.isalpha() or c == " "
			symbols_to_discard = [ key for key in self.keys() if not valid_symbol(key) ]
			self.discard(symbols_to_discard)


	def score(self, content):
		if not self.case_sensitive:
			content = content.lower()

		return super(EnglishAnalyzer, self).score(content)


import glob, sys, operator
from freqens.analyzer import Analyzer


def get_files_matching_patterns(patterns):
	return reduce(operator.add, (glob.glob(pattern) for pattern in patterns), [])

def main():
	patterns = sys.argv[1:]
	files = get_files_matching_patterns(patterns)

	analyzer = Analyzer()

	for raw_file in files:
		analyzer.feed_from_raw_file(raw_file)

	print analyzer.serialize()


from collections import Counter
from heapq import nlargest

def calculate_counter_size(other):
	return sum(other.values())

class NormalizedCounter(object):
	"""
	Like collection.Counter but performs relative counts
	"""

	def __init__(self, content = None):	
		"""Constructor for the NormalizedCounter class

		:param content: the string from which the counter will be initially constructed
		"""

		self.counter = Counter(content)
		self.absolute_size = calculate_counter_size(self.counter)

	def most_common(self, n = None):
		most_common_abs = self.counter.most_common(n)

		return [ (char, 1.*freq / self.absolute_size ) for (char, freq) in most_common_abs ]

	def elements(self):
		return self.counter.iterkeys()

	def insert(self, iterable):
		self.absolute_size += len(iterable)
		self.counter.update(iterable)

	def absolute_counts(self):
		return dict( self.counter )

	def transform(self, transformation):
		"""
		Maps the character distribution to another character distribution changing the keys through
		the transformation function provided

		For example: { "a": 0.25, "A" : 0.75 } gets mapped to { "A" : 1 } with transformation = lambda s: s.upper()
		"""
		new_counter = Counter()
		for key in self:
			new_key = transformation(key)
			new_counter[new_key] += self.counter[key]

		self.counter = new_counter

	def __add__(self, other):
		""" Performs an absolute sum of two NormalizedCounters """
		result = NormalizedCounter()
		if isinstance(other, Counter):
			result.absolute_size = calculate_counter_size(other) + self.absolute_size
			result.counter = other + self.counter
		
		elif isinstance(other, NormalizedCounter):
			result.absolute_size = other.absolute_size + self.absolute_size
			result.counter = other.counter + self.counter
		else:
			raise TypeError("other must be either a Counter or a NormalizedCounter")

		return result

	

	def __len__(self):
		""" Returns the number of distinct elements """
		return len(self.counter.keys())
	
	def __getitem__(self, key):
		return (self.counter[key] * 1.) / self.absolute_size

	def __delitem__(self, key):
		self.absolute_size -= self.counter[key]
		del self.counter[key]

	def __iter__(self):
		return self.counter.__iter__()	# an iterator over the keys

	def iteritems(self):
		return ( (key, self[key]) for key in self )

	def __contains__(self, key):
		return self[key] > 0

	def __str__(self):
		result_dict = { key: self[key] for key in self.counter }

		return result_dict.__str__()
